import 'dart:mirrors';

import 'package:annotation_crawler/annotation_crawler.dart';
import 'package:test/test.dart';

part 'src/definitions.dart';

main() {
  group("annotatedDeclarations()", () {
    test("if passed an instance of an annotation, should get any top level declarations with that instance", () {
      expect(annotatedDeclarations(AnnotationInstance).map((decl) => decl.declaration.simpleName), [#foo]);
    });
    test("if passed an annotation type, should get any top level declarations with an instance of that type", () {
      final decls = annotatedDeclarations(ClassAnnotation);

      expect(decls.map((decl) => decl.declaration.simpleName),
          unorderedEquals([#ClassWithAnnotation, #SecondClassWithAnnotation]));
      expect(decls.map((decl) => decl.annotation), everyElement(const ClassAnnotation()));
    });

    test("if passed an instance of an annotation, should get all declarations in the given scope with the declaration",
        () {
      final decls = annotatedDeclarations(AnnotationInstance, on: reflectClass(ClassWithMethods));
      expect(decls.map((decl) => decl.declaration.simpleName), [#bar]);
    });
    test("if passed an annotation type, should get all declarations in scope with an instance of that type", () {
      final decls = annotatedDeclarations(MethodAnnotation, on: reflectClass(ClassWithMethods));
      expect(decls.map((decl) => decl.declaration.simpleName),
          unorderedEquals([#methodWithAnnotation, #anotherMethodWithAnnotation, #variable, #annotatedImplementation]));
      expect(decls.map((decl) => decl.annotation), everyElement(const MethodAnnotation()));
    });

    test("if recursive is true, annotations on super classes and interfaces are included", () {
      final decls = annotatedDeclarations(MethodAnnotation, on: reflectClass(ClassWithMethods), recursive: true);
      expect(
          decls.map((decl) => decl.declaration.simpleName),
          unorderedEquals([
            #methodWithAnnotation,
            #anotherMethodWithAnnotation,
            #variable,
            #annotatedImplementation,
            #interfaceMethod,
            #superclassVariable
          ]));
      expect(decls.map((decl) => decl.declaration).whereType<MethodMirror>().fold(true, (r, m) => r && !m.isAbstract),
          true);
    });
  });

  group('findClasses()', () {
    test("returns all classes with provided annotation class", () {
      final foundClasses = findClasses(ClassAnnotation);

      expect(foundClasses.length, equals(2));

      expect(
          foundClasses.firstWhere((classMirror) => classMirror.reflectedType == ClassWithAnnotation,
                  orElse: () => null) !=
              null,
          equals(true));

      expect(
          foundClasses.firstWhere((ClassMirror classMirror) => classMirror.reflectedType == SecondClassWithAnnotation,
                  orElse: () => null) !=
              null,
          equals(true));
    });
    test("doesn't return classes with unnused annotation", () {
      final foundClasses = findClasses(UnusedAnnotation);

      expect(foundClasses.length, equals(0));
    });
  });

  group('findMethodsOnInstance()', () {
    test("returns all methods provided annotation class", () {
      final instance = new ClassWithMethods();

      final foundMethods = findMethodsOnInstance(instance, MethodAnnotation);

      expect(foundMethods.map((m) => m.simpleName),
          [#methodWithAnnotation, #anotherMethodWithAnnotation, #annotatedImplementation]);
    });
    test("doesn't return methods with unused annotation", () {
      // ignore: unused_local_variable
      final cls = reflectClass(ClassWithMethods);

      final instance = new ClassWithMethods();

      final foundMethods = findMethodsOnInstance(instance, UnusedAnnotation);

      expect(foundMethods, []);
    });
  });

  group('findMethodsOnClass()', () {
    test("returns all methods provided annotation class", () {
      final foundMethods = findMethodsOnClass(ClassWithMethods, MethodAnnotation);

      expect(foundMethods.map((m) => m.simpleName),
          [#methodWithAnnotation, #anotherMethodWithAnnotation, #annotatedImplementation]);
    });
    test("doesn't return methods with unused annotation", () {
      final foundMethods = findMethodsOnClass(ClassWithMethods, UnusedAnnotation);

      expect(foundMethods, []);
    });
  });
}
